package org.apache.bcel.classfile;

import java.io.DataOutputStream;
import java.util.Arrays;
import java.io.IOException;
import java.io.DataInput;

public final class Module extends Attribute
{
    private final int moduleNameIndex;
    private final int moduleFlags;
    private final int moduleVersionIndex;
    private ModuleRequires[] requiresTable;
    private ModuleExports[] exportsTable;
    private ModuleOpens[] opensTable;
    private final int usesCount;
    private final int[] usesIndex;
    private ModuleProvides[] providesTable;
    
    Module(final int nameIndex, final int length, final DataInput input, final ConstantPool constantPool) throws IOException {
        super((byte)22, nameIndex, length, constantPool);
        this.moduleNameIndex = input.readUnsignedShort();
        this.moduleFlags = input.readUnsignedShort();
        this.moduleVersionIndex = input.readUnsignedShort();
        final int requiresCount = input.readUnsignedShort();
        this.requiresTable = new ModuleRequires[requiresCount];
        for (int i = 0; i < requiresCount; ++i) {
            this.requiresTable[i] = new ModuleRequires(input);
        }
        final int exportsCount = input.readUnsignedShort();
        this.exportsTable = new ModuleExports[exportsCount];
        for (int j = 0; j < exportsCount; ++j) {
            this.exportsTable[j] = new ModuleExports(input);
        }
        final int opensCount = input.readUnsignedShort();
        this.opensTable = new ModuleOpens[opensCount];
        for (int k = 0; k < opensCount; ++k) {
            this.opensTable[k] = new ModuleOpens(input);
        }
        this.usesCount = input.readUnsignedShort();
        this.usesIndex = new int[this.usesCount];
        for (int k = 0; k < this.usesCount; ++k) {
            this.usesIndex[k] = input.readUnsignedShort();
        }
        final int providesCount = input.readUnsignedShort();
        this.providesTable = new ModuleProvides[providesCount];
        for (int l = 0; l < providesCount; ++l) {
            this.providesTable[l] = new ModuleProvides(input);
        }
    }
    
    @Override
    public void accept(final Visitor v) {
        v.visitModule(this);
    }
    
    @Override
    public Attribute copy(final ConstantPool constantPool) {
        final Module c = (Module)this.clone();
        Arrays.setAll(c.requiresTable = new ModuleRequires[this.requiresTable.length], i -> this.requiresTable[i].copy());
        Arrays.setAll(c.exportsTable = new ModuleExports[this.exportsTable.length], i -> this.exportsTable[i].copy());
        Arrays.setAll(c.opensTable = new ModuleOpens[this.opensTable.length], i -> this.opensTable[i].copy());
        Arrays.setAll(c.providesTable = new ModuleProvides[this.providesTable.length], i -> this.providesTable[i].copy());
        c.setConstantPool(constantPool);
        return c;
    }
    
    @Override
    public void dump(final DataOutputStream file) throws IOException {
        super.dump(file);
        file.writeShort(this.moduleNameIndex);
        file.writeShort(this.moduleFlags);
        file.writeShort(this.moduleVersionIndex);
        file.writeShort(this.requiresTable.length);
        for (final ModuleRequires entry : this.requiresTable) {
            entry.dump(file);
        }
        file.writeShort(this.exportsTable.length);
        for (final ModuleExports entry2 : this.exportsTable) {
            entry2.dump(file);
        }
        file.writeShort(this.opensTable.length);
        for (final ModuleOpens entry3 : this.opensTable) {
            entry3.dump(file);
        }
        file.writeShort(this.usesIndex.length);
        for (final int entry4 : this.usesIndex) {
            file.writeShort(entry4);
        }
        file.writeShort(this.providesTable.length);
        for (final ModuleProvides entry5 : this.providesTable) {
            entry5.dump(file);
        }
    }
    
    public ModuleExports[] getExportsTable() {
        return this.exportsTable;
    }
    
    public ModuleOpens[] getOpensTable() {
        return this.opensTable;
    }
    
    public ModuleProvides[] getProvidesTable() {
        return this.providesTable;
    }
    
    public ModuleRequires[] getRequiresTable() {
        return this.requiresTable;
    }
    
    @Override
    public String toString() {
        final ConstantPool cp = super.getConstantPool();
        final StringBuilder buf = new StringBuilder();
        buf.append("Module:\n");
        buf.append("  name:    ").append(Utility.pathToPackage(cp.getConstantString(this.moduleNameIndex, (byte)19))).append("\n");
        buf.append("  flags:   ").append(String.format("%04x", this.moduleFlags)).append("\n");
        final String version = (this.moduleVersionIndex == 0) ? "0" : cp.getConstantString(this.moduleVersionIndex, (byte)1);
        buf.append("  version: ").append(version).append("\n");
        buf.append("  requires(").append(this.requiresTable.length).append("):\n");
        for (final ModuleRequires module : this.requiresTable) {
            buf.append("    ").append(module.toString(cp)).append("\n");
        }
        buf.append("  exports(").append(this.exportsTable.length).append("):\n");
        for (final ModuleExports module2 : this.exportsTable) {
            buf.append("    ").append(module2.toString(cp)).append("\n");
        }
        buf.append("  opens(").append(this.opensTable.length).append("):\n");
        for (final ModuleOpens module3 : this.opensTable) {
            buf.append("    ").append(module3.toString(cp)).append("\n");
        }
        buf.append("  uses(").append(this.usesIndex.length).append("):\n");
        for (final int index : this.usesIndex) {
            final String className = cp.getConstantString(index, (byte)7);
            buf.append("    ").append(Utility.compactClassName(className, false)).append("\n");
        }
        buf.append("  provides(").append(this.providesTable.length).append("):\n");
        for (final ModuleProvides module4 : this.providesTable) {
            buf.append("    ").append(module4.toString(cp)).append("\n");
        }
        return buf.substring();
    }
}
